#include "web_config.h"

namespace config {
  iotwebconf::TextTParameter<STRING_LEN> stringParam = iotwebconf::Builder<iotwebconf::TextTParameter< STRING_LEN>>(
      "stringParam").label("String param").build();
}

namespace web_config {
#if defined(ESP8266)
  ESP8266WebServer http(80); // Create a webserver object that listens for HTTP request on port 80
#elif defined(ESP32)
  WebServer http(80);
#endif

  DNSServer dnsServer; //TODO: Check if needed

  IotWebConf *iotWebConf;

// -- Configuration specific key. The value should be modified if config structure was changed.
  const char config_version[] = "ampel_test_v4";

  static const char chooserValues[][STRING_LEN] = { "red", "blue", "darkYellow" };
  static const char chooserNames[][STRING_LEN] = { "Red", "Blue", "Dark yellow" };

  iotwebconf::ParameterGroup serviceParams = iotwebconf::ParameterGroup("Services", "Services");

  iotwebconf::CheckboxTParameter ampelWifiParam = iotwebconf::Builder<iotwebconf::CheckboxTParameter>("WiFi").label(
      "WiFi").defaultValue(true).build();
  iotwebconf::CheckboxTParameter ampelCSVParam =
      iotwebconf::Builder<iotwebconf::CheckboxTParameter>("CSV").label("CSV").defaultValue(true).build();
  iotwebconf::CheckboxTParameter ampelMQTTParam = iotwebconf::Builder<iotwebconf::CheckboxTParameter>("MQTT").label(
      "MQTT").defaultValue(true).build();
  iotwebconf::CheckboxTParameter ampelLoRaWANParam =
      iotwebconf::Builder<iotwebconf::CheckboxTParameter>("LoRaWAN").label("LoRaWAN").defaultValue(false).build();

  iotwebconf::ParameterGroup co2Params = iotwebconf::ParameterGroup("co2", "CO2 Sensor");

  //TODO: UIntTParameter?
  iotwebconf::IntTParameter<uint16_t> timestepParam = iotwebconf::Builder<iotwebconf::IntTParameter<uint16_t>>(
      "timestep").label("Measurement timestep").defaultValue(MEASUREMENT_TIMESTEP).min(2).max(1800).step(1).placeholder(
      "2..1800").build();

  iotwebconf::FloatTParameter temperatureOffsetParam =
      iotwebconf::Builder<iotwebconf::FloatTParameter>("temp_offset").label("Temperature offset").defaultValue(
      TEMPERATURE_OFFSET).placeholder("-3").build();

  iotwebconf::ParameterGroup group2 = iotwebconf::ParameterGroup("c_factor", "Calibration factor");
  iotwebconf::CheckboxTParameter checkboxParam =
      iotwebconf::Builder<iotwebconf::CheckboxTParameter>("checkParam").label("Check param").defaultValue(true).build();
  iotwebconf::SelectTParameter<STRING_LEN> chooserParam =
      iotwebconf::Builder<iotwebconf::SelectTParameter< STRING_LEN>>("chooseParam").label("Choose param").optionValues(
          (const char*) chooserValues).optionNames((const char*) chooserNames).optionCount(
          sizeof(chooserValues) / STRING_LEN).nameLength(STRING_LEN).build();

  void update() {
    iotWebConf->doLoop(); // Listen for HTTP requests from clients
  }

  void setWifiConnectionCallback(void (*function)()) {
    iotWebConf->setWifiConnectionCallback(function);
  }

  void setWifiConnectionFailedCallback(void (*function)()) {
    iotWebConf->setWifiConnectionFailedHandler([&function]() -> iotwebconf::WifiAuthInfo* {
      function();
      return NULL;
    });
  }

  void initialize() {
    iotWebConf = new IotWebConf(ampel.sensorId, &dnsServer, &http, HTTP_PASSWORD, config_version);

    serviceParams.addItem(&ampelWifiParam);
    serviceParams.addItem(&ampelMQTTParam);
    serviceParams.addItem(&ampelCSVParam);
    serviceParams.addItem(&ampelLoRaWANParam);

    co2Params.addItem(&timestepParam);
    co2Params.addItem(&temperatureOffsetParam);

    iotWebConf->addParameterGroup(&serviceParams);
    iotWebConf->addParameterGroup(&co2Params);

    sensor_console::defineCommand("reset_config", []() {
      Serial.println(F("Resetting config..."));
      iotWebConf->getSystemParameterGroup()->applyDefaultValue();
      iotWebConf->saveConfig();
      Serial.println(F("Done!"));
    }, F("(resets the complete IotWeb config)"));

    sensor_console::defineStringCommand("conf", [](char *new_value) {
      Serial.print(F("Setting stringParam to "));
      Serial.println(new_value);
      strncpy(config::stringParam.value(), new_value, STRING_LEN);
      Serial.println(F("Done"));
      iotWebConf->saveConfig();
    }, F("some_text (config setter)"));

#if !defined(AMPEL_WIFI)
    iotWebConf->loadConfig();
    return;
#endif

    const int ONBOARD_LED_PIN = 2;
# ifdef ESP8266
    iotWebConf->setStatusPin(ONBOARD_LED_PIN, LOW);
# else
    iotWebConf->setStatusPin(ONBOARD_LED_PIN, HIGH);
# endif
    iotWebConf->setWifiConnectionTimeoutMs(1000UL * WIFI_TIMEOUT);

#if defined(ESP8266)
  WiFi.hostname(ampel.sensorId);
#elif defined(ESP32)
    WiFi.setHostname(ampel.sensorId);
#endif

    iotWebConf->skipApStartup();
    //TODO: Add callbacks
    //TODO: Add LED effects
    //TODO: Allow offline config loading
    //TODO: Add default values for SSID/password
    //TODO: Add other params
    //TODO: Use HTTP_USER / HTTP_PASSWORD for config
    //TODO: Move to own class
    //TODO: Remove AP Password config?
    //TODO: Save LoRaWAN key if possible?
    //FIXME: Why does MQTT fail? (on ESP32)

    iotWebConf->init();

    //TODO: Authenticate only if required?
    //TODO: / captive fast return?

    http.on("/config", [] {
      iotWebConf->handleConfig();
    });

    http.onNotFound([]() {
      iotWebConf->handleNotFound();
    });
  }
}
